// -*- mode:C++ -*-
//Header:
//Title: Sample of Soccer Client Program
//File: client.C
//Author: NODA, Itsuki
//Date: 1996/05/19
//EndHeader:
 
//History:
// This file is created originally.
//	by NODA, Itsuki [1996/06/06] (Ver. 1.0)
//
// Change to chheck the argument of sqrt to avoid domain error.
//	by NODA, Itsuki [1996/06/07] (Ver. 1.1)
//
//EndHistory:

//Usage:
// client [ options ] TEAMNAME [ PLAYSTRYLE ]
//  options: 
//	-h HOSTNAME 
//	-p PORTNUMBER
//  PLAYSTYLE:
//	dango ; only chases the ball
//      nao ; original type
//      nao_midfielder ; original midfielder type
//      nao_defender ; original defender type
//	defender ; keeps the position between the ball and theown goal 
//	midfielder ; chases the ball if the ball is in the midfield,
//		     else surrounds in the midfield. 
//	left_wingback ; covers the left side.
//	right_wingback ; covers the right side.
//	forward ; waits the ball in the opponent-team-side.
//	top ; waits the ball in front of the opponent goal.
//	okano ; dash!! dash!! dash!!
//	takeda ; ambush
//	jo ; ambush
//	maezono ; dribble!! dribble!! dribble!!
//	michiki ; left wing back
//	narahashi ; right wing back
//	nakata ; passer
//	hattori, ito, endo ; midfielder
//	tanaka, suzuki, matsuda ; defender
//	kawaguchi ; goalie
//	GK, LCB, RCB, LWB, RWB, BR, OMF, LW, RW, LFW, RFW
//EndUsage:

//======================================================================
//Headers
//======================================================================
#include <String.h>
#include <iostream.h>
#include <strstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <GetOpt.h>

#include "udpif.h"
#include "client.h"

//======================================================================
//Main
//======================================================================
int main(int argc, char **argv) {
     Player player ;

     if(!(player.init(argc, argv))) exit(-1) ;

     player.mainloop() ;

     player.close() ;
}


//======================================================================
//Top Facilities
//======================================================================
Boolean Player::init(int argc, char **argv) {
//Analyze Args
     mygetopt(argc,argv) ;
  
     sock = init_connection((char*)((const char*)serverhost),serverport) ;
     if(sock.socketfd == -1) {
	  cerr << "Can't open socket." << endl ;
	  return False ;
     }

     ostrstream ost(send_buffer,bufsize) ;

     ost << "(init " << team << ")" << endl << ends ;

     if(send_message(send_buffer,sock) == -1) {
	  cerr << "Can't connect to the server." << endl ;
	  return False ;
     }
     
//Receive Init message
     Int n ;
     while(!(n = receive_message(recv_buffer, bufsize, &sock))) {} ;
     if(n == -1) {
	  cerr << "????" << endl ;
	  return False ;
     }
     Int l ;
     char mode[BUFFERSIZE] ;
     char sidename[NAMEMAX] ;
     if(3 > sscanf(recv_buffer,"(init %[lr] %d %[^)])%n", 
		   sidename, &unum, mode, &l)) {
	  cerr << "Refused by the server: " << recv_buffer << endl ; 
	  return False ;
     }

// Setup Side
     side = str_to_Side(sidename) ;

// Setup Land Mark
     String* LMStr = ((side == Side_L)?LandMarkStrL:LandMarkStrR) ;
     for(Int i = 0 ; i < LM_max ; ++i) {
	  landmark[i] = findVisualObject(LMStr[i]) ;
	  landmark[i]->pos = LandMarkPos[i] ;
	  landmark[i]->fixp = True ;
     }
     ball = landmark[LM_ball] ;
     goal = landmark[LM_goal] ;
     owngoal = landmark[LM_owngoal] ;	
     ball->fixp = False ;

// Move to initial Position
     move(play_style) ;

// Final Report
     cout << "Connect with " << serverhost << " [" << serverport << "]" 
	  << endl ;
     cout << "  side: " << side << endl ;
     cout << "  unum: " << unum << endl ;
     cout << "  mode: " << mode << endl ; 
     cout << "  play_style: " << PlayStyleStr[play_style] << endl ;

     return True ;
}

void Player::mainloop() {
     start_sensor() ;

     switch(play_style) {
     case PS_dango:
	  play_as_dango() ; break ;
     case PS_nao:
          play_as_nao() ;break ;
     case PS_nao_midfielder:
          play_as_nao_midfielder() ;break ;
     case PS_nao_defender:
          play_as_nao_defender() ;break ;
     case PS_defender:
	  play_as_defender() ; break ;
     case PS_midfielder:
	  play_as_midfielder() ; break ;
     case PS_top:
	  play_as_forward() ; break ;
     case PS_forward:
	  play_as_forward2() ; break ;
     case PS_left_wingback:
	  play_as_midfielder2(PPos(-20,-15),30,20) ; break ;
     case PS_right_wingback:
	  play_as_midfielder2(PPos(-20, 15),30,20) ; break ;
     case PS_okano:
	  play_as_okano() ; break ;
     case PS_takeda:
	  play_as_takeda() ; break ;
     case PS_jo:
	  play_as_jo() ; break ;
     case PS_maezono:
	  play_as_maezono() ; break ;
     case PS_michiki:
	  play_as_michiki() ; break ;
     case PS_narahashi:
	  play_as_narahashi() ; break ;
     case PS_nakata:
	  play_as_nakata() ; break ;
     case PS_hattori:
	  play_as_hattori() ; break ;
     case PS_ito:
	  play_as_ito() ; break ;
     case PS_endo:
	  play_as_endo() ; break ;
     case PS_tanaka:
	  play_as_tanaka() ; break ;
     case PS_suzuki:
	  play_as_suzuki() ; break ;
     case PS_matsuda:
	  play_as_matsuda() ; break ;
     case PS_kawaguchi:
	  play_as_kawaguchi() ; break ;

     case PS_GK: {
	  PPos dst[] = {PM_LWB, PM_RWB} ;
	  play_as_relayer(PM_GK, PPos(5,5), dst, 2) ; }
	  break ;
     case PS_LCB: {
	  PPos dst[] = {PM_LWB, PM_BR} ;
	  play_as_relayer(PM_LCB, PPos(25,15), dst, 2) ;}
	  break ;
     case PS_RCB: {
	  PPos dst[] = {PM_RWB, PM_BR} ;
	  play_as_relayer(PM_RCB, PPos(25,15), dst, 2) ;}
	  break ;
     case PS_LWB: {
	  PPos dst[] = {PM_LW, PM_OMF} ;
	  play_as_relayer(PM_LWB, PPos(45,12), dst, 2) ;}
	  break ;
     case PS_RWB: {
	  PPos dst[] = {PM_RW, PM_OMF} ;
	  play_as_relayer(PM_RWB, PPos(45,12), dst, 2) ;}
	  break ;
     case PS_BR: {
	  PPos dst[] = {PM_LW, PM_OMF, PM_RW} ;
	  play_as_relayer(PM_BR, PPos(20,35), dst, 3) ;}
	  break ;
     case PS_OMF: {
	  PPos dst[] = {PM_LW, PM_LCFW, PM_RCFW, PM_RW} ;
	  play_as_relayer(PM_OMF, PPos(20,35), dst, 4) ;}
	  break ;
     case PS_LW: {
	  PPos dst[] = {PM_GL, PM_LCFW} ;
	  play_as_relayer(PM_LW, PPos(40,12), dst, 2) ;}
	  break ;
     case PS_RW: {
	  PPos dst[] = {PM_GL, PM_RCFW} ;
	  play_as_relayer(PM_RW, PPos(40,12), dst, 2) ;}
	  break ;
     case PS_LCFW: {
	  PPos dst[] = {PM_GL, PM_GLR, PM_RCFW} ;
	  play_as_relayer(PM_LCFW, PPos(15,10), dst, 3) ;}
	  break ;
     case PS_RCFW: {
	  PPos dst[] = {PM_GL, PM_GLL, PM_LCFW} ;
	  play_as_relayer(PM_RCFW, PPos(15,10), dst, 3) ;}
	  break ;
     default:
	  exit(-1) ;
     }
}

void Player::close() {
     close_connection(this->sock) ;
}

//======================================================================
//Initializations
//======================================================================
void Player::mygetopt(int argc, char** argv) {
     GetOpt getopt(argc,argv,"h:p:") ;
     Int option_char ;
     while ((option_char = getopt ()) != EOF) {
	  switch (option_char) {
	  case 'h': serverhost = getopt.optarg; break;
	  case 'p': serverport = atoi(getopt.optarg); break;
	  case '?': usage() ; exit(-1) ; 
	       }
     }

     for(Int i = 0, j = getopt.optind ; j < argc ; i++, j++) {
	  switch(i) {
	  case 0:
	       team = argv[j] ; break ;
	  case 1:
	       play_style_str = argv[j] ; break ;
	  default:
	       usage() ; exit(-1) ;
	  }
     }
     
     play_style = str_to_position(play_style_str) ;
     if(play_style < 0) { usage() ; exit(-1) ; }
}     
     
void usage() {
     cerr << 
"Usage:\n\t\% client [options] TEAMNAME [ PLAYSTYLE ]
	options: 
		-h HOSTNAME 
		-p PORTNUMBER
	PLAYSTYLE:
		dango
                nao
                nao_midfielder
                nao_defender
		defender
		midfielder
		left_wingback
		right_wingback
		forward
		top
		okano,	takeda	jo
		maezono
		michiki narahashi
		nakata
		hattori, ito, endo 
		tanaka, suzuki, matsuda 
		kawaguchi
" << ends ;

}

//======================================================================
//Sensor Process
//======================================================================
Player* P ; 

void sensor_cycle(void) { P->check_message() ; }

void Player::check_message() {
     clock++ ; 
     Int r = receive_message(recv_buffer,bufsize,&sock) ;
     switch(r) {
     case 0: return ;
     case 1: analyze_message() ; break ;
     default: 
	  cerr << "Warning[" << __LINE__
	       << "]:trouble in receive_message." << endl ;
     }
}

void Player::start_sensor() {
     
     cout << "Sensor process started!!" << endl ;
     
     struct itimerval itv ;

     itv.it_interval.tv_sec = 0 ;
     itv.it_interval.tv_usec = RECV_INTERVAL_MILISEC * 1000 ;
     itv.it_value.tv_sec = 0 ;
     itv.it_value.tv_usec = RECV_INTERVAL_MILISEC * 1000 ;

     setitimer(ITIMER_REAL, &itv, NULL) ;
     P = this ;
     signal(SIGALRM, (void (*)(int))sensor_cycle) ;
}

void Player::analyze_message() {
     Int p,l ;

     p = 0 ;
     static char tmp[NAMEMAX] ;
     if(1 > sscanf(&(recv_buffer[p]),"(%s %n",tmp,&l)) {
	  cerr << "Warning[" << __LINE__ << "]:Illegal message:" 
	       << recv_buffer << endl ;
	  return ;
     }
     p += l ;
 
     Int t ;
     if(!strcmp(tmp,"see")) {
	  if(1 > sscanf(&(recv_buffer[p]),"%d %n",&t,&l)) {
	       cerr << "Warning[" << __LINE__ << ": Illegal message:" 
		    << recv_buffer << endl ;	
	       return ;
	  }
	  p += l ;
	  vtime = time = t ;

	  Real dist, dir ;
	  Int vdir ;
	  Int c ;

	  while(recv_buffer[p] != ')') {
	       if(1 > sscanf(&(recv_buffer[p])," ((%[^)]) %n",tmp,&l)) {
		    cerr << "Warning[" << __LINE__ << "]: Illegal message:" 
			 << &(recv_buffer[p]) << endl ;
		    return ;
	       }
	       p += l ;
	       	
	       if(3 == (c = sscanf(&(recv_buffer[p]),"%lf %lf %d)%n",
			      &dist,&dir,&vdir,&l))) {
	       } else if(2 == (c = sscanf(&(recv_buffer[p]),"%lf %lf)%n",
					  &dist,&dir,&l))) {
	       } else if(1 == (c = sscanf(&(recv_buffer[p]),"%lf)%n",
					  &dir,&l))) {
	       } else {
		    cerr << "Warning[" << __LINE__ << "]: Illegal message:" 
			 << recv_buffer << endl ;
		    return ;
	       }
	       p += l ;

	       renewVisualObject(tmp,dist,dir,vdir,c) ;
	  }
     } else if (!strcmp(tmp,"hear")) {
	  Real dir ;
	  if(2 == sscanf(&(recv_buffer[p]),"%d %lf %n",&t,&dir,&l)) {
	       p += l ;
	       atime = time = t ;
	       cout << "Message:" << dir << ":" 
		    << &(recv_buffer[p]) << endl ;
	  } else if (2 == sscanf(&(recv_buffer[p]),"%d %s %n",&t,tmp,&l)) {
	       p += l ;
	       atime = time = t ;
	       cout << "Message:" << tmp << ":"
		    << &(recv_buffer[p]) << endl ;
	  } else {
	       cerr << "Warning[" << __LINE__ << "]: Illegal message:" 
		    << recv_buffer << endl ;	
	       return ;
	  }
     } else {
	  cerr << "Warning[" << __LINE__ << "]: Illegal message:" 
	       << recv_buffer << endl ;	
	  return ;
     }
}

//======================================================================
//Void Action
//======================================================================
Int ptime = 0 ;
void Player::voidaction() {
     Int i ;
     turn(10) ;
     spause() ;
     i = time - ptime ;
     if(!(0 == i)) {
	  showStatus() ;
	  ptime = time ;
     }
}

//======================================================================
//Send Command
//======================================================================
PlayStyle Player::str_to_position(String PosName) {
     for(Int i = 0 ; i < PS_max ; i++) {
	  if(PosName == PlayStyleStr[i]) {
	       return (PlayStyle)i ;
	  }
     }
     return (PlayStyle)-1 ;
}

void Player::move(PlayStyle pos) {
     Real x,y ;

     randomize() ;
     
     x = drand(PlayStyleRange[pos][0],PlayStyleRange[pos][1]) ;
     y = drand(PlayStyleRange[pos][2],PlayStyleRange[pos][3]) ;

     move(x,y) ;
}

void Player::move(Real x, Real y) {

     cout << "Move:(" << x << "," << y << ")" << endl ;

     sprintf(send_buffer,"(move %d %d)",(Int)x,(Int)y) ;
     send_message(send_buffer,sock) ;
     
}

void Player::turn(Real moment) {
     if(moment < EPS && moment > -EPS) { return ;} ;

//     cout << "Turn:" << moment << endl ;

     moment = normalize_angle(moment) ;
     sprintf(send_buffer,"(turn %d)",(Int)moment) ;
     send_message(send_buffer,sock) ;
     cdir += moment ;
     cdir = (Int)normalize_angle(cdir) ;

     npause() ;
}

void Player::turn_to(Real adir) {
     turn(adir - cdir) ;
}

void Player::dash(Real power) {

//     cout << "Dash:" << power << endl ;

     sprintf(send_buffer,"(dash %d)",(Int)power) ;
     send_message(send_buffer,sock) ;

     npause() ;
}

void Player::kick(Real power, Real dir) {

     cout << "Kick:" << power << "," << dir << endl ;

     sprintf(send_buffer,"(kick %d %d)",(Int)power, (Int)dir) ;
     send_message(send_buffer,sock) ;

     npause() ;
}

void Player::say(Buffer msg) {
     
     cout << "Say:" << msg << endl ; 

     sprintf(send_buffer,"(say %s)",msg) ;
     send_message(send_buffer,sock) ;

     npause() ;
}

void Player::change_view(ViewAngle va, ViewQuality vq) {

     npause() ;

     cout << "ChangeView: " << va << "," << vq << endl ; 

     sprintf(send_buffer,"(change_view %s %s)",
	     (const char*)ViewAngleStr[va],(const char*)ViewQualityStr[vq]);
     send_message(send_buffer,sock) ;
     cva = va ; 
     cvq = vq ;

     npause() ;
}

//======================================================================
//Visual Object Manipuration
//======================================================================
VisualObject* Player::findVisualObject(String name) {
     Int i ;
     for(i = 0 ; i < N_vobj ; i++) {
	  if(vobj[i].name == name) {
	       return &(vobj[i]) ;
	  }
     }
     N_vobj++ ;
     vobj[i].name = name ;
     vobj[i].vdir = 0 ;
     vobj[i].dir = 0 ;
     vobj[i].dist = 0 ;
     vobj[i].time = -1 ;
     vobj[i].pos.x = vobj[i].pos.y = 0.0 ;
     vobj[i].fixp = False ;
     return &(vobj[i]) ;
}

void Player::renewVisualObject(String name, 
			       Real dist,
			       Real dir,
			       Int vdir,
			       Int type) {
     VisualObject* o = findVisualObject(name) ;

     o->time = time ;

     switch(type) {
     case 3: 
	  o->dist = dist ;
	  o->dir = dir + cdir ;
	  o->vdir = vdir ;
	  break ;
     case 2:
	  if(cvq == VQ_High) {
	       o->dist = dist ;
	       o->dir = dir + cdir;
	  } else {
	       o->dir = dist + cdir;
	       o->vdir = (Int)dir ;
	  }
	  break ;
     case 1:
	  o->dir = dir + cdir;
	  break ;
     }
}

void Player::showStatus() {
     cout << "[" << time << "]--------------------------------" 
	  << endl;
     Int i ;
     for(i = 0 ; i < N_vobj ; i++) {
	  cout << "  #" << i << ": " << vobj[i].name << "[" << vobj[i].time 
	       << "]:Rh(" 
	       << vobj[i].dist << "," << vobj[i].dir << ")," 
	       << vobj[i].vdir 
	       << ", XY(" << vobj[i].pos.x << "," << vobj[i].pos.y << ")"
	       <<endl ;
     }
}

//======================================================================
//Check ViewInfo
//======================================================================
void Player::wait_new_vinfo() {
     Int tm = time ;
     //while(tm >= vtime) spause() ; 
     while(tm >= vtime) ;
}

Boolean Player::new_vinfo_p(VisualObject* obj) {
     return (Boolean)(time - obj->time < 10) ;
}
     

//======================================================================
//Basic Actions(Turns)
//======================================================================

void Player::look_around() {
     cout << "Now, looking around." << endl ;
     //change_view(VA_Wide,VQ_Low) ;
     wait_new_vinfo() ;
     for(Int i = 0 ; i < 4 ; i++) {
	  turn(80) ;
	  wait_new_vinfo() ;
     }
}

void Player::look_around_half() {
     cout << "Now, looking around (half)." << endl ;
     //change_view(VA_Wide,VQ_Low) ;
     wait_new_vinfo() ;
     for(Int i = 0 ; i < 2 ; i++) {
	  turn(-90) ;
	  wait_new_vinfo() ;
     }
}

void Player::look_at(String name) {
     look_at(findVisualObject(name)) ;
}

void Player::look_at(VisualObject* obj) {
     cout << "Now, looking at " << obj->name << "." << endl ;
     //change_view(VA_Narrow,VQ_High) ;

     turn_to(obj->dir) ;
     Int tm = time ;
     wait_new_vinfo() ;
     if(obj->time <= tm) {
	  cout << "!!! Lose site of " << obj->name << "." << endl ;
	  search_obj(obj) ;
     }
}

void Player::turn_to(String name) {
     turn_to(findVisualObject(name)) ;
}
void Player::turn_to(VisualObject* obj) {
     look_at(obj) ;
     turn_to(obj->dir) ;
}

void Player::search_obj(VisualObject* obj) {
     cout << "Now, searching " << obj->name << "." << endl ;
     Int tm = time ;
     wait_new_vinfo() ;
     while(obj->time <= tm) {
	  turn(80) ;
	  wait_new_vinfo() ;
     }
     turn_to(obj->dir) ;
}

//======================================================================
//Basic Actions(Chase)
//======================================================================
void Player::chase(String name, Real maxpow, Real offset) {
     chase(findVisualObject(name), maxpow, offset) ;
}
void Player::chase(VisualObject* obj, Real maxpow, Real offset) {
     turn_to(obj) ;

     Real d = find_clear_way(0.0,obj->dist) ;
     turn(d) ;

     Dash(obj->dist,maxpow, offset) ;
     _pause(2) ;
}


void Player::go_toward(PPos tpos, Real maxpow, Real offset) {
     estimate_position(False) ;
     Real dx = tpos.x - pos.x, dy = tpos.y - pos.y ;
     Real adir = Atan2(dy, dx) ;
     turn(adir - edir) ;

     Real dist = sqrt(dx * dx + dy * dy) ;

     Real d = find_clear_way(0.0,dist) ;
     turn(d) ;
     
     Dash(dist,maxpow,offset) ;
}
     

//======================================================================
//Basic Actions(Kick)
//======================================================================
Boolean Player::kickable() {	
     return (Boolean)((ball->dist < 2.15) && new_vinfo_p(ball)) ;
} 

void Player::shoot_to_goal() {
     kick_to(goal) ;
     dash(5);
}

void Player::kick_to(VisualObject* obj, Real maxpow, Real offset) {
     cout << "Now, kicking the ball to " << obj->name << "." << endl ;
     turn_to(obj) ;
     Kick(obj->dist,0.0,maxpow,offset) ;
}

void Player::kick_immediately(VisualObject* obj, Real maxpow, Real offset) {
     cout << "Now, kicking (immediately) the ball to " 
       << obj->name << "." << endl ;
     PPos tpos = estimate_object_pos(obj) ;
     kick_immediately(tpos, maxpow, offset) ;
}

void Player::kick_immediately(PPos tpos, Real maxpow, Real offset) {
     estimate_position(False) ;
     Real dx = tpos.x - pos.x, dy = tpos.y - pos.y ;
     Real adir = Atan2(dy, dx) ;
     Real ddir = adir - edir ;
     Real dist = sqrt(dx * dx + dy * dy) ;

     Kick(dist, ddir, maxpow, offset) ;
     turn(ddir) ;
}

//======================================================================
//Estimate Current Position
//======================================================================
void Player::estimate_position(Boolean check_old_p) {
     VisualObject *lm1, *lm2 ;
     
     Boolean oldp = False ;

     do {
	  if(landmark[LM_goal]->time > landmark[LM_owngoal]->time) {
	       lm1 = landmark[LM_goal] ;
	       lm2 = landmark[LM_owngoal] ;
	  } else {
	       lm2 = landmark[LM_goal] ;
	       lm1 = landmark[LM_owngoal] ;
	  }

	  for(Int i = LM_owngoal+1 ; i < LM_max ; i++) {
	       if(landmark[i]->time > lm1->time) {
		    lm2 = lm1 ;
		    lm1 = landmark[i] ;
	       } else if (landmark[i]->time > lm2->time) {
		    lm2 = landmark[i] ;
	       }
	  }
	  if (time - lm2->time > 10 && check_old_p) {
	       oldp = True ;
	       look_around_half() ;
	  } else { oldp = False ; }
     } while (oldp) ; 

     Real x1 = lm1->pos.x, y1 = lm1->pos.y, r1 = lm1->dist ;
     Real x2 = lm2->pos.x, y2 = lm2->pos.y, r2 = lm2->dist ;
     Boolean revp = (Boolean)(y1 == y2) ;
     if(revp) {
	  Real b ;
	  b = x1 ; x1 = y1 ; y1 = b ;
	  b = x2 ; x2 = y2 ; y2 = b ;
     }

     Real A = (x1 * x1 - x2 * x2) + (y1 * y1 - y2 * y2) 
	  - (r1 * r1 - r2 * r2) ;
     Real B = -2 * (x1 - x2) ;
     Real C = A / (2 * (y1 - y2)) ;
     Real D = B / (2 * (y1 - y2)) ;
     Real E = 1 + D * D ;
     Real F = -x1 + C * D - D * y1 ;
     Real G = x1 * x1 + C * C - 2 * C * y1 + y1 * y1 - r1 * r1 ;
     Real H0 = F * F - E * G ;
     Real H = ((H0 < 0.0)? 0.0 : H0) ;
     Real x01 = (- F + sqrt(H)) / E ;
     Real x02 = (- F - sqrt(H)) / E ;
     Real y01 = C + D * x01 ;
     Real y02 = C + D * x02 ;

     Real theta1 = Atan2(x1 - x01, y1 - y01) 
	  - Atan2(x2 - x01, y2 - y01) ; 
     //     Real theta2 = Atan2(x1 - x02, y1 - y02) 
     //		   - Atan2(x2 - x02, y2 - y02) ; 
     theta1 = normalize_angle(theta1) ;
     
     Real theta0 = normalize_angle(-(lm1->dir - lm2->dir)) ;
     
     
     if(theta0 * theta1 > 0) {
	  pos.x = x01 ;
	  pos.y = y01 ;
     } else {
	  pos.x = x02 ;
	  pos.y = y02 ;
     }
     
     if(revp) {
	  Real b;
	  b = pos.x ; pos.x = pos.y ; pos.y = b ;
     }
     
     Real tau1 = Atan2(y1 - pos.y,x1 - pos.x) ;
     Real tau01 = (lm1->dir - cdir) ;
     edir = normalize_angle((tau1 - tau01)) ;
     
}


//======================================================================
//Estimate Object Position
//======================================================================
PPos Player::estimate_object_pos(VisualObject* obj) {
     if(obj->fixp) {
	  return obj->pos ;
     } else {
          estimate_position(False) ;
	  obj->pos.x = pos.x + 
	       obj->dist * Cos(edir + obj->dir - cdir) ;
	  obj->pos.y = pos.y + 
	       obj->dist * Sin(edir + obj->dir - cdir) ;
	  return obj->pos ;
     }
}


//======================================================================
//Dango Soccer
//======================================================================
void Player::play_as_dango() {
     look_around() ;
     while(True) {
	  estimate_position() ;
          dash(100) ;
	  if(kickable()) {
	       shoot_to_goal() ;
          } else {
	       chase(ball) ;
          }
     }
}	  

//======================================================================
//Nao Soccer
//======================================================================
void Player::play_as_nao() {
       look_around() ;
       estimate_position() ;
       PPos bpos = estimate_object_pos(ball) ;
       while(True) {
          if((abs((Int)pos.y) < 20.0) && (pos.x < 35.0)) {
              estimate_position() ;
              bpos = estimate_object_pos(ball) ;
	  } else {
                chase_2(ball) ;
                kick_to(goal, 100, 50) ;
	  }
          if(kickable()) {
              if(pos.x > -15.0) {
                     if(goal->dist < 30.0) {
                         if(bpos.x < pos.x) chase_2(ball) ;
                         else      kick_to(goal, 100, 50) ;
                      } else           offensive_pass() ;
	      } else                active_pass() ;
	  } else { 
                if((ball->dist) < 25.0) {
                    if(pos.x > 20.0) chase_3(ball) ;
                    else           chase_2(ball) ;
		} else chase_2(ball) ;
	  }
     }
}


//=====================================================================
//Nao Midfielder
//=====================================================================
void Player::play_as_nao_midfielder() {
     look_around() ;
     estimate_position() ;
     PPos bpos = estimate_object_pos(ball) ;
     while(True) {
	  if((abs((Int)pos.y) < 15.0) && (pos.x < 30.0)){
              estimate_position() ;
              bpos = estimate_object_pos(ball) ;
	  } else {
                chase_2(ball) ;
                kick_to(goal, 100, 50) ;
	  }
	  if(kickable()) {
               if(pos.x > -15.0) {
                     if(goal->dist < 30.0) {
                          if(bpos.x < pos.x) chase_2(ball) ;
                          else      kick_to(goal, 100, 50) ;
                     } else           offensive_pass() ;
	      } else                active_pass() ;
	  } else if(ball->dist < 10.0 && nearest_to_ballp()) {
	       chase_2(ball) ;
	  } else if(bpos.x < pos.x) {
	       turn_to(owngoal) ; dash(100) ;
	  } else if(ball->dist < 30.0 && nearest_to_ballp()) {
	       chase_3(ball) ;
	  } else {
	       wandering(PPos(0.0,0.0),30.0) ;
	  }
     }
}


//=====================================================================
//Nao Defender
//=====================================================================
void Player::play_as_nao_defender() {
     look_around() ;
     estimate_position() ;
     PPos bpos = estimate_object_pos(ball) ;
     while(True) { 
          if((abs((Int)pos.y) < 20.0) && (pos.x < 35.0)) {
              estimate_position() ;
              bpos = estimate_object_pos(ball) ;
	  } else {
                chase_2(ball) ;
                kick_to(goal, 100, 50) ;
	  }
          if(kickable()) {
              if(pos.x > -15.0) offensive_pass() ;
              else              active_pass() ;
	  } else { 
                if(((ball->dist) < 20.0)||(bpos.x < -25.0)) {
                    if(pos.x > 20.0) chase_3(ball) ;
                    else           chase_2(ball) ;
		} else {
                      while((pos.x > -35.0) && (pos.x > (bpos.x - 35.0))) {
                              estimate_position() ;
                              bpos = estimate_object_pos(ball) ;
                              turn_to(owngoal) ;
                              dash(100) ;
		      }
                      keep_defensive_pos() ;
		}
	  }
     }
}


//======================================================================
//Nao Function
//=====================================================================
void Player::chase_2(VisualObject* obj) {
     
     turn_to(obj) ;
     Real d = find_clear_way(0.0,obj->dist) ;
     turn(d) ; dash(5) ;
     PPos bpos = estimate_object_pos(ball) ;
     if((obj->dist) > 5.0) {
             dash(100) ;
     } else {
           if(((obj->dist) < 3.0) && (pos.x > bpos.x)) {
//               if(((bpos.y - 0.1) < pos.y) && (pos.y < (bpos.y + 0.1))) {
//                    turn(45) ; dash(50) ;
//                    turn_to(owngoal) ; dash(70) ;
//                    turn(90) ; kick(10.0, 0.0) ;     
//               } else {
//                     turn_to(owngoal) ;
//                     dash(50) ;
//               }
               if(kickable()) {
                   if(pos.y > 0) turn(-120) ; 
                   else          turn(120) ;
               } kick(10.0, 0.0) ;                   
           } else {
                 dash(50) ;
	   }
     }
}


void Player::chase_3(VisualObject* obj) {
     
     turn_to(obj) ;
     Real d = find_clear_way(0.0,obj->dist) ;
     turn(d) ; dash(5) ;
     PPos bpos = estimate_object_pos(ball) ;
 
     if((obj->dist) > 5.0) {
           dash(100) ;
     } else {
           if(((obj->dist) < 3.0) && (abs((Int)pos.y) < abs((Int)bpos.y))) {
//               if(((bpos.x - 0.1) < pos.x) && (pos.x < (bpos.x + 0.1))) {
//                    turn(45) ;  dash(60) ;
//                    turn(-45) ; dash(60) ;
//                    turn(45) ; kick(10.0, 0.0) ;
//               }
               if(kickable()) {
                    if(pos.y > 0) turn(100) ; 
                    else          turn(-100) ;
               kick(10.0, 0.0) ;
               }  
	   } else {
                 dash(50) ;
	   }
     }
}


void Player::chase_4(VisualObject* obj) {
     
     turn_to(obj) ;
     dash(50) ;
}


void Player::active_pass() {
     VisualObject* target = newest_fore_teammate() ;
   
     if((target == (VisualObject*)NULL)||((target->dist) < 15.0)
                                       ||(time - target->time > 30)) {
	        if(pos.y > 0) target = landmark[LM_flag_cr] ;
	        else	      target = landmark[LM_flag_cl] ;
     }
     
     kick_immediately(target) ;
}

void Player::offensive_pass() {
     VisualObject* target = newest_fore_teammate() ;

     if((target == (VisualObject*)NULL)||(time - target->time > 30)
        ||(target->dist < 10.0)||(target->dist > 20.0)) {
                
        target = goal ;
     }
     if(target = goal) power_dribble_to(target) ;
     else kick_immediately(target) ;
}

void Player::power_dribble_to(VisualObject* obj) {
     cout << "Dribble!! to " << obj->name << endl ;
     turn_to(obj) ;
     kick(50.0,0.0) ;
     dash(100.0) ;
}

//======================================================================
//Defender
//======================================================================
void Player::play_as_defender() {
     look_around() ;
     while(True) {
	  estimate_position() ;
          dash(100) ;                //Speed_Up!!
	  if(kickable()) {
	       if(pos.x < -45.0 && abs((Int)pos.y) < 10.0) {
		     cout << "Clear!!" << endl ;
		     kick_immediately(landmark[LM_flag_cr],70,40);
	       } else {
		     clear_or_pass() ;
	       }
	  } else if(ball->dist < 10.0) {
	       chase(ball) ;
	  } else {
	       keep_defensive_pos() ;
	  }
     }
}

void Player::keep_defensive_pos() {
     PPos bpos = estimate_object_pos(ball) ;
     PPos ogpos = estimate_object_pos(owngoal) ;
     PPos tpos(0.8 * ogpos.x + 0.2 * bpos.x,
	       0.8 * ogpos.y + 0.2 * bpos.y) ;
     go_toward(tpos) ;
     npause() ;
     turn_to(ball) ;
}

Boolean Player::teammatep(VisualObject *obj) {
     int p = -1 ;
     char b[MINIBUFSIZE] ;
     if(1 > sscanf(obj->name,"player %s%n",b,&p)) {
	  return False ;
     }
//     if(str_to_Side(b) == side) {
     if(team == b) {
	  return True ;
     } else { 
	  return False ;
     }
}

Boolean Player::playerp(VisualObject *obj) {
     int p = -1 ;
     sscanf(obj->name,"player%n",&p) ;
     if(p > 0) 
	  return True ;
     else
	  return False ;
}

Real Player::find_clear_way(Real ddir,Real dist) {
     Real w = 5.0 ;
     while((!clearwayp(ddir,dist))&&( w < 180 )) { 
	  ddir += drand(-w,w) ; w += 5.0 ; 
     }
     if(ddir != 0.0) ;
//     cout << "find_clear_way(" << ddir << ")[" << w << "]" << endl ;
     return ddir ;
}
     
Boolean Player::clearwayp(Real to_dir,Real to_dist) {
     estimate_position(False) ;
     for(Int i = 0 ; i < N_vobj ; i++ ) {
	  VisualObject* obj = &(vobj[i]) ;
	  if(playerp(obj)) {
	       if(time - obj->time < 10) {
		    if( to_dist > obj->dist ) {
			 Real d = obj->dist * Sin(to_dir + edir - obj->dir) ;
			 if(d < 1.0) {
			      return False ;
			 }
		    }
	       }
	  }
     }
     return True ;
}

VisualObject* Player::newest_fore_teammate() {
     VisualObject* mate = (VisualObject*)NULL ;
     for(Int i = 0; i < N_vobj ; i++) {
	  VisualObject* obj = &(vobj[i]) ;
	  if(teammatep(obj)) {
	       PPos tpos = estimate_object_pos(obj) ;
	       if(tpos.x > pos.x) {
		    if(mate == (VisualObject*)NULL) {
			 mate = obj ;
		    } else if (obj->time > mate->time) {
			 mate = obj ;
		    } 
	       }
	  }
     }
     return mate ;
}

void Player::clear_or_pass() {
     VisualObject* target = newest_fore_teammate() ;

     if((target == (VisualObject*)NULL)||(time - target->time > 100)) {
	  if(pos.y > 0) target = landmark[LM_flag_cr] ;
	  else	        target = landmark[LM_flag_cl] ;
     }
     
     kick_immediately(target) ;
}

//======================================================================
//MidFielder
//======================================================================
Boolean Player::nearest_to_ballp() {
     PPos bpos = estimate_object_pos(ball) ;
     for(Int i = 0 ; i < N_vobj ; i++) {
	  VisualObject* obj = &(vobj[i]) ;
	  if(teammatep(obj)) {
	       if(time - obj->time < 10) {
		    PPos ppos = estimate_object_pos(obj) ;
		    Real d = bpos.distance(ppos) ;
		    if(d < ball->dist) 
			 return False ;
	       }
	  }
     }
     return True ;
}

void Player::play_as_midfielder() {
     look_around() ;
     while(True) {
	  estimate_position() ;
          dash(100) ;                 //Speed_Up!!
	  PPos bpos = estimate_object_pos(ball) ;
	  if(kickable()) {
	       shoot_or_pass() ;
	  } else if(ball->dist < 10.0 && nearest_to_ballp()) {
	       chase(ball) ;
	  } else if(bpos.x < pos.x) {
	       go_toward(owngoal->pos) ;
	  } else if(ball->dist < 30.0 && nearest_to_ballp()) {
	       chase(ball) ;
	  } else {
	       wandering(PPos(0.0,0.0),30.0) ;
	  }
     }
}

void Player::shoot_or_pass() {
     if(goal->dist < 20.0) {
	  shoot_to_goal() ;
     } else {
	  VisualObject* target = newest_fore_teammate() ;

	  if((target == (VisualObject*)NULL)||(time - target->time > 100)) {
	       target = goal ;
	  }
	  kick_immediately(target) ;
     }
}

void Player::wandering(PPos center, Real radius) {
     cout << "Now, wandering." << endl ;

//     Real dx = center.x - pos.x ;
//     Real dy = center.y - pos.y ;
//     Real d = sqrt(dx * dx + dy * dy) ;
     cout << center << pos << endl ;
     Real d = center.distance(pos) ;
     
     if(d < radius) {
	  Real tdir = drand(-180,180) ;
	  turn(tdir) ;
	  Real pow = drand(10,15) ;
	  dash(pow) ;
	  lpause() ;
     } else {
	  go_toward(center) ;
     }
     turn_to(ball) ;
}

	  
//======================================================================
//Forward
//======================================================================
void Player::play_as_forward() {
     look_around() ;
     while(True) {
	  estimate_position() ;
          dash(100) ;                            //Speed_Up!!
	  PPos bpos = estimate_object_pos(ball) ;
	  if(kickable()) {
	       shoot_to_goal() ;
	  } else if(ball->dist < 20.0) {
	       chase(ball) ;
	  } else if(bpos.x > pos.x) {
	       chase(ball) ;
	  } else {
	       wandering(landmark[LM_flag_pfc]->pos,20.0) ;
	  }
     }
}

void Player::play_as_forward2() {
     look_around() ;
     while(True) {
	  estimate_position() ;
	  PPos bpos = estimate_object_pos(ball) ;
	  if(kickable()) {
	       shoot_to_goal() ;
	  } else if(ball->dist < 30.0) {
	       chase(ball) ;
	  } else if(bpos.x > pos.x) {
	       chase(ball) ;
	  } else {
	       wandering(PPos(10,0),30.0) ;
	  }
     }
}

//======================================================================
//MidFielder2
//======================================================================
void Player::wandering2(PPos center, Real rx, Real ry) {
     cout << "Now, wandering." << endl ;
     
     Real d = center.distance(pos,PPos(rx,ry)) ;
     
     if(d < 1.0) {
	  Real tdir = drand(-180,180) ;
	  turn(tdir) ;
	  Real pow = drand(10,20) ;
	  dash(pow) ;
	  lpause() ;
     } else {
	  go_toward(center) ;
     }
     turn_to(ball) ;
}

void Player::play_as_midfielder2(PPos center, Real rx, Real ry) {
     look_around() ;
     while(True) {
	  estimate_position() ;
	  PPos bpos = estimate_object_pos(ball) ;
	  if(kickable()) {
	       shoot_or_pass() ;
	  } else if( center.distance(pos,PPos(rx,ry)) < 1.0 ) {
	       if(nearest_to_ballp()) 
		    chase(ball) ;
	       else 
		    wandering(bpos,10.0) ;
	  } else {
	       wandering2(center,rx,ry) ;
	  }
     }
}
	  
			  
//======================================================================
//Okano!!
//======================================================================
void Player::okano_dash(VisualObject* obj) {
     cout << "Go Go Okano!! Dash to " << obj->name << endl ;
     turn_to(obj) ;
     for(Int i = -1 ; i < (Int)(obj->dist/5.0) ; i++) {
       dash(30) ; 
       spause() ;
     }
}

void Player::play_as_okano() {
     look_around() ;
     while(True) {
	  estimate_position() ;
	  PPos bpos = estimate_object_pos(ball) ;
	  if(kickable()) {
	    cout << "Shoot!!" << endl ; 
	       shoot_to_goal();
	       turn_to(goal) ;
	       dash(30);
	  } else if(bpos.x > pos.x) {
	       okano_dash(ball) ;
	  } else {
	    cout << "Wandering!!" << endl ; 
	       wandering(PPos(0.0,0.0),10.0) ;
	  }
     }
}

//======================================================================
//takeda
//======================================================================
void Player::play_as_takeda() {
     look_around() ;
     while(True) {
	  estimate_position() ;
	  PPos bpos = estimate_object_pos(ball) ;
	  if(kickable()) {
	    cout << "Shoot!!" << endl ;
	       shoot_to_goal() ;
	  } else if(ball->dist < 10.0) {
	    cout << "Chase the ball!!" << endl ;
	       chase(ball) ;
	  } else if(bpos.x > pos.x) {
	    cout << "Chase the ball!!" << endl ;
	       chase(ball) ;
	  } else {
	    cout << "Wandering!!" << endl ;
	       wandering(landmark[LM_flag_pfc]->pos,10.0) ;
	  }
     }
}

//======================================================================
//Jo
//======================================================================
void Player::play_as_jo() {
     look_around() ;
     while(True) {
	  estimate_position() ;
	  look_at(ball);
	  PPos bpos = estimate_object_pos(ball) ;
	  if(kickable()) {
	    cout << "Shoot!!" << endl ;
	    shoot_to_goal() ;
	  } else if(ball->dist < 5.0) {
	    cout << "Chase the ball!!" << endl ;
	    chase(ball,50,1) ;
	  } else if(bpos.x > pos.x - 10) {
	    cout << "Rush to goal!!" << endl ;
	    PPos tpos = landmark[LM_flag_pfc]->pos ;
	    tpos.x += 4.0 ; 
	    go_toward(tpos,40,20) ;
	  } else {
	    cout << "Wandering!!" << endl ;
	    wandering(PPos(10.0,0.0),10.0) ;
	  }
     }
}

//======================================================================
//Maezono!!
//======================================================================
void Player::dribble_to(VisualObject* obj) {
     cout << "Dribble!! to " << obj->name << endl ;
     turn_to(obj) ;
     kick(10.0,0.0) ;
     dash(20.0) ;
     wait_new_vinfo();
}

void Player::play_as_maezono() {
     look_around() ;
     while(True) {
	  estimate_position() ;
	  PPos bpos = estimate_object_pos(ball) ;
	  if(kickable()){
	       if(abs((Int)normalize_angle(ball->dir - goal->dir)) < 120){
		 if(goal->dist < 20) {
		   cout << "Shoot!!" << endl ;
		   shoot_to_goal() ;
		 } else {
		   dribble_to(goal) ;
		 }
	       } else {
		 cout << "Turning with the ball." << endl ;
		 kick(4.0,90.0) ;
		 turn(90.0) ;
	       }
//	  } else if(ball->dist < 3.0 && ball->pos.x < pos.x ) {
//	       look_at(ball) ;
//	       turn(-80) ;
//	       dash(10) ;
	  } else if(ball->dist < 20.0) {
	       chase(ball) ;
	  } else if(bpos.x > pos.x) {
	       chase(ball) ;
	  } else {
	       wandering(PPos(-10.0,0.0),10.0) ;
	  }
     }
}


//======================================================================
//Michiki!!
//======================================================================

void Player::centering(Real power) {
     if(pos.y > 10.0) 
       kick_to(landmark[LM_flag_fl],power,30) ;
     else if (pos.y < -10.0)
       kick_to(landmark[LM_flag_fr],power,30) ;
     else 
       kick_to(goal,power,30) ;
}

void Player::play_as_michiki() {
     look_around() ;
     while(True) {
	  estimate_position() ;
	  PPos bpos = estimate_object_pos(ball) ;
	  if(kickable()){
	       if(abs((Int)normalize_angle(ball->dir - goal->dir)) < 120){
		 if(pos.x > 30){
		   cout << "Centering!!" << endl ;
		   centering(70.0) ;
		 } else {
		   cout << "Dribbling!!" << endl ;
		   dribble_to(landmark[LM_flag_fl]) ;
		 }
	       } else {
		 cout << "Turning with the ball." << endl ;
		 kick(4.0,90.0) ;
		 turn(90.0) ;
	       }
	  } else if(ball->dist < 10.0) {
	       cout << "Chasing the ball!!" << endl ;
	       chase(ball) ;
	  } else if(bpos.x > pos.x && bpos.y < -10.0) {
	       cout << "Chasing the ball for offence!!" << endl ;
	       chase(ball) ;
	  } else if(bpos.x < pos.x && bpos.y < 0.0) {
	       cout << "Chase the ball for defence!!" << endl;
	       chase(ball) ;
	  } else if(bpos.x < pos.x) {
	       cout << "Go back to my position!!" << endl;
	       chase(landmark[LM_flag_pbl]) ;
	  } else {
	       cout << "Wandering!!" << endl ;
	       wandering(PPos(-30.0,-20.0),10.0) ;
	  }
     }
}

//======================================================================
//Narahashi
//======================================================================

void Player::play_as_narahashi() {
     look_around() ;
     while(True) {
	  estimate_position() ;
	  PPos bpos = estimate_object_pos(ball) ;
	  if(kickable()){
	       if(abs((Int)normalize_angle(ball->dir - goal->dir)) < 120){
		 if(pos.x > 30){
		   cout << "Centering!!" << endl ;
		   centering(70.0) ;
		 } else {
		   cout << "Dribbling!!" << endl ;
		   dribble_to(landmark[LM_flag_fr]) ;
		 }
	       } else {
		 cout << "Turning with the ball." << endl ;
		 kick(4.0,-90.0) ;
		 turn(-90.0) ;
	       }
	  } else if(ball->dist < 10.0) {
	       cout << "Chasing the ball!!" << endl ;
	       chase(ball) ;
	  } else if(bpos.x > pos.x && bpos.y > 10.0) {
	       cout << "Chasing the ball for offence!!" << endl ;
	       chase(ball) ;
	  } else if(bpos.x < pos.x && bpos.y > 0.0) {
	       cout << "Chase the ball for defence!!" << endl;
	       chase(ball) ;
	  } else if(bpos.x < pos.x) {
	       cout << "Go back to my Position!!" << endl ;
	       chase(landmark[LM_flag_pbr]) ;
	  } else {
	       cout << "Wandering!!" << endl ;
	       wandering(landmark[LM_flag_pbr]->pos,10.0) ;
	  }
     }
}

//======================================================================
//Nakata
//======================================================================
void Player::through_pass() {
  VisualObject* target = newest_fore_teammate() ;

  if((target == (VisualObject*)NULL)||(time - target->time > 50)) {
    centering(100) ;
  } else {
    PPos tpos = estimate_object_pos(target) ;
    tpos.x += 10.0 ; 
    kick_immediately(tpos,70,30) ;
  }
}

void Player::play_as_nakata() {
     look_around() ;
     while(True) {
	  estimate_position() ;
	  PPos bpos = estimate_object_pos(ball) ;
	  if(kickable()) {
	    cout << "Pass!!" << endl ;
	       through_pass() ;
	  } else if(ball->dist < 20.0 && nearest_to_ballp()) {
	     cout << "Chase the ball!!" << endl ;
	       chase(ball) ;
	  } else if(bpos.x < pos.x + 10.0) {
	     cout << "Come back home." << endl ; 
	       go_toward(owngoal->pos) ;
	  } else {
	     cout << "Wandering." << endl ; 
	       wandering(PPos(-10.0,0.0),15.0) ;
	  }
     }
}

//======================================================================
//Kawaguchi
//======================================================================
void Player::play_as_kawaguchi() {
     look_around() ;
     while(True) {
	  estimate_position() ;
	  PPos bpos = estimate_object_pos(ball) ;
	  if(kickable()) {
	       if(pos.x < -45.0 && abs((Int)pos.y) < 10.0) {
		     cout << "Clear!!" << endl ;
		     kick_immediately(landmark[LM_flag_cr],70,40);
	       } else {
		     cout << "Clear or Pass!!" << endl ;
		     clear_or_pass() ;
	       }
	  } else if(ball->dist < 10.0) {
	       cout << "Chase the ball!!" << endl ; 
	       chase(ball,20.0,3.0) ;
	       look_at(ball) ;
	  } else {
	       if(pos.distance(owngoal->pos) > 1.0)
		 if(bpos.x < pos.x || pos.x < owngoal->pos.x) {
		   cout << "Back to owngoal hastily!!" << endl ;  
		   go_toward(owngoal->pos) ;
		 } else {
		   cout << "back to goal backward!!" << endl ;
		   go_toward_backward(owngoal->pos) ;
		 }
	       turn_to(ball) ;
	  }
     }
}

void Player::go_toward_backward(PPos tpos) {
     estimate_position(False) ;
     Real dx = tpos.x - pos.x, dy = tpos.y - pos.y ;
     Real adir = Atan2(dy, dx) ;
     turn(adir - edir + 180) ;

     Real dist = sqrt(dx * dx + dy * dy) ;
     
     Dash(-dist,-10,-10) ;
}

//======================================================================
//General
//======================================================================
void Player::play_as_relayer(PPos center, 
			     PPos r, 
			     PPos* dest, 
			     Int Ndist) {
  
  look_around() ;
  while(True) {
    estimate_position() ;
    look_at(ball) ;
    PPos bpos = estimate_object_pos(ball) ;
    if(kickable()) {
      Int i = irand(Ndist) ;
      cout << "Pass to " << dest[i] << endl ;
      kick_immediately(dest[i],70,30) ;
    } else if (center.distance(bpos,r) < 1.0) {
      cout << "Ball is at " << bpos << ". This is my area." << endl ;
      cout << "Chase the ball!!" << endl ;
      chase(ball,30.0,10.0) ;
    } else if (center.distance(pos) > 1.0) {
      cout << "Return to my position " << center << "." << endl ;
      go_toward(center,40,15) ;
      npause() ;
    } else {
      cout << "Waiting." << endl ;
      if(irand(2)) turn(90) ;
      else         turn(-90) ;
    }
  }
}

       
